mod worker;

use crate::worker::*;
use im::HashMap;
use libpulse_binding::context::introspect::{SinkInfo, SinkInputInfo, SourceInfo};
use libpulse_binding::volume::{ChannelVolumes, Volume};
use std::num::NonZeroU32;
use std::sync::Arc;
use std::time::Instant;

#[allow(unused_imports)]
use libpulse_binding::mainloop::api::Mainloop as _;

pub type EntityId = NonZeroU32;

#[derive(Debug, Eq, PartialEq, Copy, Clone)]
pub enum EntityKind {
    Source,
    Sink,
    SinkInput,
}

#[derive(Debug, Eq, PartialEq, Clone)]
pub enum EntityMetadata {
    Source {
        name: String,
        description: String,
    },
    Sink {
        name: String,
        description: String,
    },
    SinkInput {
        description: String,
        binary_name: Option<String>,
        application_name: Option<String>,
    },
}

#[derive(Debug, Clone, PartialEq)]
pub struct EntityState {
    id: EntityId,
    is_muted: bool,
    channel_volumes: ChannelVolumes,
    metadata: EntityMetadata,
}

impl EntityState {
    pub fn id(&self) -> &EntityId {
        &self.id
    }

    pub fn is_muted(&self) -> bool {
        self.is_muted
    }

    pub fn channel_volumes(&self) -> Vec<f32> {
        self.channel_volumes.get().iter().map(|v| v.0 as f32 / Volume::NORMAL.0 as f32).collect()
    }

    pub fn metadata(&self) -> &EntityMetadata {
        &self.metadata
    }

    pub fn kind(&self) -> EntityKind {
        match &self.metadata {
            EntityMetadata::Source { .. } => EntityKind::Source,
            EntityMetadata::Sink { .. } => EntityKind::Sink,
            EntityMetadata::SinkInput { .. } => EntityKind::SinkInput,
        }
    }
}

impl From<&SourceInfo<'_>> for EntityState {
    fn from(value: &SourceInfo) -> Self {
        EntityState {
            id: NonZeroU32::new(value.index).unwrap(),
            is_muted: value.mute,
            channel_volumes: value.volume,
            metadata: EntityMetadata::Source {
                name: value.name.clone().unwrap_or_default().into_owned(),
                description: value.description.clone().unwrap_or_default().into_owned(),
            },
        }
    }
}

impl From<&SinkInfo<'_>> for EntityState {
    fn from(value: &SinkInfo) -> Self {
        EntityState {
            id: NonZeroU32::new(value.index).unwrap(),
            is_muted: value.mute,
            channel_volumes: value.volume,
            metadata: EntityMetadata::Sink {
                name: value.name.clone().unwrap_or_default().into_owned(),
                description: value.description.clone().unwrap_or_default().into_owned(),
            },
        }
    }
}

impl From<&SinkInputInfo<'_>> for EntityState {
    fn from(value: &SinkInputInfo) -> Self {
        EntityState {
            id: NonZeroU32::new(value.index).unwrap(),
            is_muted: value.mute,
            channel_volumes: value.volume,
            metadata: EntityMetadata::SinkInput {
                description: value.name.clone().unwrap_or_default().into_owned(),
                application_name: value
                    .proplist
                    .get("application.name")
                    .map(|v| String::from_utf8_lossy(v).trim_end_matches(char::from(0)).to_owned()),
                binary_name: value
                    .proplist
                    .get("application.process.binary")
                    .map(|v| String::from_utf8_lossy(v).trim_end_matches(char::from(0)).to_owned()),
            },
        }
    }
}

#[derive(Debug, Clone)]
pub struct State {
    timestamp: Instant,
    default_sink_id: Option<EntityId>,
    default_source_id: Option<EntityId>,
    entities_by_id: HashMap<EntityId, Arc<EntityState>>,
}

impl State {
    pub fn timestamp(&self) -> &Instant {
        &self.timestamp
    }

    pub fn default_sink_id(&self) -> &Option<EntityId> {
        &self.default_sink_id
    }

    pub fn default_source_id(&self) -> &Option<EntityId> {
        &self.default_source_id
    }

    pub fn entities_by_id(&self) -> &HashMap<EntityId, Arc<EntityState>> {
        &self.entities_by_id
    }
}

#[derive(Debug, Clone)]
pub struct PulseAudioVolumeInterface {
    worker: Arc<PaWorker>,
}

impl PulseAudioVolumeInterface {
    pub fn new(client_name: String) -> PulseAudioVolumeInterface {
        PulseAudioVolumeInterface {
            worker: Arc::new(PaWorker::spawn(client_name)),
        }
    }

    pub fn subscribe_to_state(&self) -> (Arc<State>, async_broadcast::Receiver<Arc<State>>) {
        let rx = self.worker.state_tx.new_receiver();
        let state = self.current_state();
        (state, rx)
    }

    pub fn current_state(&self) -> Arc<State> {
        self.worker.current_state.load_full()
    }

    pub fn set_is_muted(&self, id: EntityId, value: bool) {
        self.worker.commands_tx.send(ThreadMessage::SetIsMuted { id, value }).unwrap()
    }

    pub fn set_channel_volumes(&self, id: EntityId, channel_volumes: impl Into<Box<[f32]>>) {
        self.worker
            .commands_tx
            .send(ThreadMessage::SetChannelVolumes {
                id,
                channel_volumes: channel_volumes.into(),
            })
            .unwrap()
    }

    pub fn set_default_entity(&self, id: EntityId) {
        self.worker.commands_tx.send(ThreadMessage::SetDefaultEntity { id }).unwrap()
    }
}
