use crate::{EntityId, EntityKind, EntityMetadata, EntityState, State};
use arc_swap::ArcSwap;
use im::HashMap;
use libpulse_binding::callbacks::ListResult;
use libpulse_binding::context::introspect::Introspector;
use libpulse_binding::context::subscribe::{Facility, InterestMaskSet};
use libpulse_binding::context::{subscribe, Context, FlagSet, State as ConnectionState};
use libpulse_binding::def::Retval;
use libpulse_binding::mainloop::api::Mainloop as _;
use libpulse_binding::mainloop::threaded::Mainloop;
use libpulse_binding::volume::Volume;
use std::num::NonZeroU32;
use std::sync::Arc;
use std::thread;
use std::time::Instant;

#[derive(Debug)]
pub(super) enum ThreadMessage {
    // emitted by the library user
    SetIsMuted {
        id: EntityId,
        value: bool,
    },
    SetChannelVolumes {
        id: EntityId,
        channel_volumes: Box<[f32]>,
    },
    SetDefaultEntity {
        id: EntityId,
    },
    Terminate {
        termination_signal_tx: oneshot::Sender<()>,
    },

    // emitted by libpulse
    LoadServerInfo,
    LoadSinkInput {
        entity_id: EntityId,
    },
    LoadSink {
        entity_id: EntityId,
    },
    LoadSource {
        entity_id: EntityId,
    },

    // emitted by handlers for the messages above
    HandleServerInfo {
        default_source_name: Option<Box<str>>,
        default_sink_name: Option<Box<str>>,
    },
    UpsertEntity {
        entity_state: Arc<EntityState>,
    },
    RemoveEntity {
        entity_id: EntityId,
    },
}

pub(super) struct PaThread<'worker> {
    mainloop: Mainloop,
    context: Context,
    introspector: Introspector,
    commands_tx: flume::Sender<ThreadMessage>,
    commands_rx: flume::Receiver<ThreadMessage>,
    state_tx: async_broadcast::Sender<Arc<State>>,
    current_state: &'worker ArcSwap<State>,
}

impl<'worker> PaThread<'worker> {
    pub(super) fn spawn(
        client_name: String,
        commands_tx: flume::Sender<ThreadMessage>,
        commands_rx: flume::Receiver<ThreadMessage>,
        state_tx: async_broadcast::Sender<Arc<State>>,
        current_state: &'worker ArcSwap<State>,
    ) {
        // SAFETY: current_state is owned by PaWorker.
        // Before PaWorker is dropped, it sends a ThreadMessage::Terminate.
        // When this thread receives this message, it stops further message processing.
        // The drop handler of PaWorker blocks until it has received confirmation that this thread will stop.
        let current_state: &'static ArcSwap<State> = unsafe { core::mem::transmute(current_state) };

        thread::spawn(move || {
            let mut mainloop = Mainloop::new().unwrap();
            let context = Context::new(&mainloop, &client_name).unwrap();
            let introspector = context.introspect();

            log::debug!("Starting the mainloop thread…");
            mainloop.start().expect("starting the mainloop never fails");

            let mut t = PaThread {
                mainloop,
                context,
                introspector,
                commands_tx,
                commands_rx,
                state_tx,
                current_state,
            };

            t.init();
            t.run();
        });
    }

    fn init(&mut self) {
        log::debug!("Initializing…");
        self.mainloop.lock();
        self.context.connect(None, FlagSet::NOFLAGS, None).unwrap();

        {
            let (context_state_change_tx, context_state_change_rx) = flume::bounded(1);
            self.context.set_state_callback(Some(Box::new(move || {
                context_state_change_tx.send(()).unwrap();
            })));

            self.mainloop.unlock();
            loop {
                context_state_change_rx.recv().unwrap();

                self.mainloop.lock();
                if self.context.get_state() == ConnectionState::Ready {
                    break;
                }
                self.mainloop.unlock();
            }
        }

        // Mainloop is still locked

        {
            let commands_tx = self.commands_tx.clone();
            self.introspector.get_sink_input_info_list(move |list_result| match list_result {
                ListResult::Error => panic!("Introspector.get_sink_input_info_list failed"),
                ListResult::End => {}
                ListResult::Item(sink_input) => commands_tx
                    .send(ThreadMessage::UpsertEntity {
                        entity_state: Arc::new(sink_input.into()),
                    })
                    .unwrap(),
            });
        }

        {
            let commands_tx = self.commands_tx.clone();
            self.introspector.get_sink_info_list(move |list_result| match list_result {
                ListResult::Error => panic!("Introspector.get_sink_info_list failed"),
                ListResult::End => {}
                ListResult::Item(sink) => commands_tx
                    .send(ThreadMessage::UpsertEntity {
                        entity_state: Arc::new(sink.into()),
                    })
                    .unwrap(),
            });
        }

        {
            let commands_tx = self.commands_tx.clone();
            self.introspector.get_source_info_list(move |list_result| match list_result {
                ListResult::Error => panic!("Introspector.get_source_info_list failed"),
                ListResult::End => {}
                ListResult::Item(source) => commands_tx
                    .send(ThreadMessage::UpsertEntity {
                        entity_state: Arc::new(source.into()),
                    })
                    .unwrap(),
            });
        }

        {
            let commands_tx = self.commands_tx.clone();
            self.context.set_subscribe_callback(Some(Box::new(move |facility, operation, entity_id| {
                let entity_id = NonZeroU32::new(entity_id).unwrap();
                let facility = facility.unwrap();

                if facility == Facility::Server {
                    commands_tx.send(ThreadMessage::LoadServerInfo).unwrap();
                } else {
                    match operation.unwrap() {
                        subscribe::Operation::Removed => {
                            commands_tx.send(ThreadMessage::RemoveEntity { entity_id }).unwrap();
                        }
                        subscribe::Operation::New | subscribe::Operation::Changed => {
                            match facility {
                                Facility::SinkInput => commands_tx.send(ThreadMessage::LoadSinkInput { entity_id }).unwrap(),
                                Facility::Sink => commands_tx.send(ThreadMessage::LoadSink { entity_id }).unwrap(),
                                Facility::Source => commands_tx.send(ThreadMessage::LoadSource { entity_id }).unwrap(),
                                _ => {}
                            };
                        }
                    };
                }
            })));
        }

        self.context.subscribe(
            InterestMaskSet::SERVER | InterestMaskSet::SINK | InterestMaskSet::SOURCE | InterestMaskSet::SINK_INPUT,
            |success| {
                if !success {
                    panic!("Context.subscribe failed")
                }
            },
        );

        self.mainloop.unlock();
    }

    fn run(mut self) {
        log::debug!("Waiting for commands…");

        'outer: loop {
            while let Ok(command) = self.commands_rx.recv() {
                self.mainloop.lock();
                let commands_tx = self.commands_tx.clone();

                match command {
                    ThreadMessage::Terminate { termination_signal_tx } => {
                        termination_signal_tx.send(()).unwrap();
                        break 'outer;
                    }
                    ThreadMessage::SetIsMuted { id, value } => {
                        if let Some(state) = self.current_state.load().entities_by_id.get(&id) {
                            match state.kind() {
                                EntityKind::Sink => self.introspector.set_sink_mute_by_index(id.into(), value, None),
                                EntityKind::Source => self.introspector.set_source_mute_by_index(id.into(), value, None),
                                EntityKind::SinkInput => self.introspector.set_sink_input_mute(id.into(), value, None),
                            };
                        }
                    }
                    ThreadMessage::SetChannelVolumes { id, channel_volumes } => {
                        if let Some(state) = self.current_state.load().entities_by_id.get(&id) {
                            let mut value = state.channel_volumes;
                            for (i, v) in channel_volumes.iter().enumerate() {
                                value.set(i as u8, Volume((Volume::NORMAL.0 as f32 * v).floor() as u32));
                            }

                            match state.kind() {
                                EntityKind::Sink => self.introspector.set_sink_volume_by_index(id.into(), &value, None),
                                EntityKind::Source => self.introspector.set_source_volume_by_index(id.into(), &value, None),
                                EntityKind::SinkInput => self.introspector.set_sink_input_volume(id.into(), &value, None),
                            };
                        }
                    }
                    ThreadMessage::SetDefaultEntity { id } => {
                        if let Some(state) = self.current_state.load().entities_by_id.get(&id) {
                            match &state.metadata {
                                EntityMetadata::Sink { name, .. } => {
                                    self.context.set_default_sink(name, |_| {});
                                }
                                EntityMetadata::Source { name, .. } => {
                                    self.context.set_default_source(name, |_| {});
                                }
                                _ => {
                                    log::error!("Setting default entity failed. Only sinks and sources can be defaults.");
                                }
                            };
                        } else {
                            log::error!("Setting default entity failed. Unknown entity id: {id}");
                        }
                    }

                    ThreadMessage::LoadServerInfo => {
                        self.introspector.get_server_info(move |server_info| {
                            commands_tx
                                .send(ThreadMessage::HandleServerInfo {
                                    default_sink_name: server_info.default_sink_name.as_ref().map(|n| n.to_string().into_boxed_str()),
                                    default_source_name: server_info.default_source_name.as_ref().map(|n| n.to_string().into_boxed_str()),
                                })
                                .unwrap();
                        });
                    }
                    ThreadMessage::LoadSinkInput { entity_id } => {
                        self.introspector.get_sink_input_info(entity_id.into(), move |list_result| match list_result {
                            ListResult::Error => panic!("Introspector.get_sink_input_info failed"),
                            ListResult::End => {}
                            ListResult::Item(sink_input) => commands_tx
                                .send(ThreadMessage::UpsertEntity {
                                    entity_state: Arc::new(sink_input.into()),
                                })
                                .unwrap(),
                        });
                    }
                    ThreadMessage::LoadSink { entity_id } => {
                        self.introspector
                            .get_sink_info_by_index(entity_id.into(), move |list_result| match list_result {
                                ListResult::Error => panic!("Introspector.get_sink_info_by_index failed"),
                                ListResult::End => {}
                                ListResult::Item(sink) => commands_tx
                                    .send(ThreadMessage::UpsertEntity {
                                        entity_state: Arc::new(sink.into()),
                                    })
                                    .unwrap(),
                            });
                    }
                    ThreadMessage::LoadSource { entity_id } => {
                        self.introspector
                            .get_source_info_by_index(entity_id.into(), move |list_result| match list_result {
                                ListResult::Error => panic!("Introspector.get_source_info_by_index failed"),
                                ListResult::End => {}
                                ListResult::Item(source) => commands_tx
                                    .send(ThreadMessage::UpsertEntity {
                                        entity_state: Arc::new(source.into()),
                                    })
                                    .unwrap(),
                            });
                    }

                    ThreadMessage::HandleServerInfo {
                        default_sink_name,
                        default_source_name,
                    } => {
                        let current_state = self.current_state.load();
                        self.set_state(Arc::new(State {
                            timestamp: Instant::now(),
                            default_source_id: default_source_name
                                .map(|search_name| {
                                    current_state
                                        .entities_by_id
                                        .iter()
                                        .find(|(_, e)| match &e.metadata {
                                            EntityMetadata::Source { name, .. } => *name == *search_name,
                                            _ => false,
                                        })
                                        .map(|(id, _)| (*id).into())
                                })
                                .flatten(),
                            default_sink_id: default_sink_name
                                .map(|search_name| {
                                    current_state
                                        .entities_by_id
                                        .iter()
                                        .find(|(_, e)| match &e.metadata {
                                            EntityMetadata::Sink { name, .. } => *name == *search_name,
                                            _ => false,
                                        })
                                        .map(|(id, _)| (*id).into())
                                })
                                .flatten(),
                            entities_by_id: current_state.entities_by_id.clone(),
                        }));
                    }
                    ThreadMessage::UpsertEntity { entity_state } => {
                        let current_state = self.current_state.load();
                        self.set_state(Arc::new(State {
                            timestamp: Instant::now(),
                            entities_by_id: current_state.entities_by_id.update(entity_state.id, entity_state),
                            ..**current_state
                        }));
                    }
                    ThreadMessage::RemoveEntity { entity_id } => {
                        let current_state = self.current_state.load();
                        self.set_state(Arc::new(State {
                            timestamp: Instant::now(),
                            entities_by_id: current_state.entities_by_id.without(&entity_id),
                            ..**current_state
                        }));
                    }
                }

                self.mainloop.unlock();
            }
        }

        // self.current_state is no longer valid beyond this point.

        self.mainloop.quit(Retval(0));
        self.mainloop.unlock();
    }

    fn set_state(&self, value: Arc<State>) {
        self.current_state.store(Arc::clone(&value));

        // If there are no subscribers, that’s ok.
        _ = self.state_tx.try_broadcast(value);
    }
}

#[derive(Debug)]
pub(super) struct PaWorker {
    pub(super) commands_tx: flume::Sender<ThreadMessage>,
    pub(super) state_tx: async_broadcast::Sender<Arc<State>>,
    pub(super) current_state: ArcSwap<State>,
}

impl PaWorker {
    pub(super) fn spawn(client_name: String) -> PaWorker {
        let (commands_tx, commands_rx) = flume::unbounded();
        let (state_tx, _) = async_broadcast::broadcast(5);

        let worker = PaWorker {
            commands_tx: commands_tx.clone(),
            state_tx: state_tx.clone(),
            current_state: ArcSwap::new(Arc::new(State {
                timestamp: Instant::now(),
                default_sink_id: None,
                default_source_id: None,
                entities_by_id: HashMap::new(),
            })),
        };

        PaThread::spawn(client_name, commands_tx, commands_rx, state_tx, &worker.current_state);

        worker
    }
}

impl Drop for PaWorker {
    fn drop(&mut self) {
        let (termination_signal_tx, termination_signal_rx) = oneshot::channel();
        self.commands_tx.send(ThreadMessage::Terminate { termination_signal_tx }).ok();
        termination_signal_rx.recv().unwrap()
    }
}
